# -----------------------------------------------------------------------------
doc="""=== Integrate region ===\n
by Evgenii Kovrigin,    08/29/05\n
Integration of one dimension in a series of spectra using the SAME rectangle. Allows 1D lineshape analysis of NMR titrations.\n
First you need to switch pointer mode to 'select' (F1). The you have to set INDEX value using [Assign Numbers] to your spectra - these numbers are used as names for plain text files (just hit [Apply] and [Close]). Integration is done by [Integrate w?] buttons.\n
'w?'is the dimension that will be REMOVED (integrated). Hitting these buttons automatically saves the plain text files from integratin of ALL the spectra in the project, numbered by INDEX values. XMGR file is also created that allow easy inspection of the traces. This data is saved in NXXX_x_w?/ folder where w? is the X-axis of the trace.\n
Data for MatLab notebook on lineshape analysis is saved separately in 'Data_for_Matlab/'. The 'name.m' files are created.Make sure to set paths in you Matlab to this folder.\n
[Append...] button saves the name of the peak(group) in the text file matlab.peaklist. These names are ready to be cut-and-paste into the MatLab notebook 'setup.m'\n

3/6/08 - part of LineShapeKin package
3/13/08 - new output (plain text) features added
"""
ver='ver. 3.0'

import sputil
import tkutil
import tkMessageBox, tkFileDialog

import curvefit
import xmgr, os
from math import *
# -----------------------------------------------------------------------------
#
#--- this is a parameter used for naming the files. Use 'es' to set it ====
INDEX_PARAMETER='relaxation_time_parameter'

dims={ 'w1':0,  # numerical order of dimensions in arrays
       'w2':1
     }


class region_rmsd_dialog(tkutil.Dialog):
  # storage for mean RMSD value
  _current_spectrum=None
  _mean_rmsd=0
  _number_of_runs=0
  def __init__(self, session):

    

    self.session = session
    self.drag_notice = None

    #--- Matlab data handling
    self.current_peak_name=''
    self.current_peak_name_saved=''
    self.matlab_path='None'

    self.my_region=None
    
    tkutil.Dialog.__init__(self, session.tk, 'LineShapeKin %s' % ver)

    self.top.columnconfigure(0, weight = 1)

    import Tkinter
    explain = ('Dragging a box defines the region to integrate')
    w = Tkinter.Label(self.top, text = explain, justify = 'left')
    w.grid(row = 0, column = 0, sticky = 'nw')

    r = Tkinter.Label(self.top, justify = 'left')
    r.grid(row = 1, column = 0, sticky = 'nw')
    self.result_label = r

    br = tkutil.button_row(self.top,
                           ('Close', self.close_cb),
                           ('Assign numbers', self.run_es),
                           ('Integrate w2', self.region_w2),
                           ('Integrate w1', self.region_w1),                           
                           ('Append peak name to MatLab peaklist', self.append_resname),                           
                           ('Help', self.help),
			   )
    br.frame.grid(row = 2, column = 0, sticky = 'nw')

    self.register_callback(1)
    self.top.bind('<Map>', lambda e, s=self: s.register_callback(1))
    self.top.bind('<Unmap>', lambda e, s=self: s.register_callback(0))



  # ---------------------------------------------------------------------------
  def help(self):
    tkMessageBox.showinfo('Integrate Regions',message=doc)
  # ---------------------------------------------------------------------------
  def run_es(self):
    self.session.command_characters('es')
  # ---------------------------------------------------------------------------
  #
  def register_callback(self, register):

    if register:
      if self.drag_notice == None:
        self.drag_notice = self.session.notify_me('drag region',
                                                  self.region_selected_cb)
    else:
      if self.drag_notice:
        self.session.dont_notify_me(self.drag_notice)
        self.drag_notice = None

  # ---------------------------------------------------------------------------
  #
  def region_selected_cb(self, region):

    if self.is_window_destroyed():
      self.register_callback(0)
      return

    #--ek
    self.my_region=region

    s = self.session.selected_spectrum()
    if s == None:
      self.result_label['text'] = 'No spectrum selected'
      return
    #print region
    rmin, rmax = region
    ranges = map(lambda vmin, vmax: '%.4g-%.4g' % (vmin, vmax), rmin, rmax)
    import string
    rtext = '[ ' + string.join(ranges, ', ') + ' ]'

    self.current_peak_name=''
    self.current_peak_name_saved=''

    self.showinfo(rtext)
        

  def showinfo(self,text):
    'to refresh info window'
    #-----------------------------------------------------------------------
    peaks=""
    for peak in self.session.selected_peaks():
      peaks+=" "+peak.assignment[:-3]
      
    msg = ('  Peaks captured    %s\n' % peaks +
           '  Region      %s\n' % text +
           '  Saved as: %s\n%s\n' %(self.current_peak_name, self.current_peak_name_saved))
    self.result_label['text'] = msg

    
#######################################################################################
#######################################################################################
#######################################################################################
#######################################################################################


  # ---------------------------------------------------------------------------
  #
  def region_w1(self):
    'integrate dimension w1'
#    print self.integrate_region(self.session.selected_spectrum(),self.my_region, 'w2')
    self.integrate_all_spectra(self.my_region, 'w2')
    
  def region_w2(self):
    'integrate dimension w2'
    #print self.integrate_region(self.session.selected_spectrum(),self.my_region, 'w1')
    self.integrate_all_spectra(self.my_region, 'w1')

  # -----------------------------------------------------------------------------
  #
  def integrate_all_spectra(self, region, X_axis):

    path=os.path.dirname(self.session.project.save_path)
    #--- make a name of peaks names ---
    peaks=""
    peak_names=""
    for peak in self.session.selected_peaks():
      peaks+=peak.assignment[:-3]+"_"
      peak_names+=peak.assignment[:-3]+" "
    peaks+="x_"+X_axis
    peaks_path=os.path.join(path,peaks)
    if not os.path.isdir(peaks_path):   os.mkdir(peaks_path)

    #-- name for a file with a list of spectra
    points_fname='points_list.txt'
    points_file=open(os.path.join(peaks_path,points_fname),'w')


    #--- paths for MatLab data ---
    matlab_datapath="Data_for_MatLab" #-- goes into a current directory
    matlab_fullpath=os.path.join(path,matlab_datapath)
    if not os.path.isdir(matlab_fullpath):   os.mkdir(matlab_fullpath)
    #--- current peak filename
    matlab_fname=peaks
    matlab_peaklist='matlab.peaklist'

    
    int_graph=xmgr.Graph(peak_names)
   
    #--- run integration
    y_maxs=[]
    y_mins=[]
    datastrings=[] #-- for matlab output
    
    for spectrum in self.session.project.spectrum_list():
      integrated, min_X, max_X, min_Y, max_Y = self.integrate_region(spectrum,region, X_axis)
      y_maxs.append(max_Y)
      y_mins.append(min_Y)
      
      #--- plain output ----
      if spectrum.saved_value(INDEX_PARAMETER):
          fname=spectrum.saved_value(INDEX_PARAMETER)
      else:
          fname=spectrum.name+".txt"
      file=open(os.path.join(peaks_path,fname),'w')

      #-- collect name of the spectrum
      points_file.write("%s\n" % fname)

      #-- collect the spectral data
      for line in integrated:
         file.write("%f\t%f\n" % (spectrum.hz_per_ppm[dims[X_axis]]*float(line[0]), line[1]))
      file.close()
      

      # ----- MatLab format output (in Roger's convention) ---
      if spectrum.saved_value(INDEX_PARAMETER):
          index=spectrum.saved_value(INDEX_PARAMETER)
      else:
          tkMessageBox.showerror('Integrate Regions',
                                 'Need INDEX value set in each spectrum!  "%s" does not have it! '%spectrum.name)
          return

      #-- datalines in Roger's system
      cur_x='x'+peaks+"Hz = [ "
      cur_y=peaks+"_%s = [ "%index
      for line in integrated:
          cur_x+=" %.6e " % (spectrum.hz_per_ppm[dims[X_axis]]*float(line[0]))
          cur_y+=" %.6e " % line[1]
          

      datastrings.append(cur_y+' ];\n')
      
    

      
      
      #-------- make XMGR file to look at -------
      int_dataset=xmgr.Dataset(integrated,spectrum.name)
      int_dataset.solid_line()
      int_dataset.no_sym()
      int_graph.add_dataset(int_dataset)

    
    int_graph.set_xaxis(X_axis,min_X, max_X)
    int_graph.set_yaxis("Data height",min(y_mins), max(y_maxs))
    int_graph.writeout(os.path.join(peaks_path,peaks)+".xmgr", legend='off')

    #-- finish building MatLab data file in Roger's convention --
    datastrings.append(cur_x+' ];\n')
    file=open(os.path.join(matlab_fullpath,matlab_fname)+'.m','w')
    for line in datastrings:
        file.write(line)
    file.close()

    #-- close point_list file
    points_file.close()

    #--- save name of exported residue to use in Matlab notebooks ---
    self.current_peak_name=matlab_fname
    self.current_peak_name_saved=''
    self.matlab_path=os.path.join(matlab_fullpath,matlab_peaklist)

    self.showinfo('')


  def append_resname(self):
    'appends the name to matlab peaklist'
    if not self.current_peak_name:
        return
    #--- save name of exported residue to use in Matlab notebooks ---
    file=open(self.matlab_path,'a')
    file.write(" '%s', " % self.current_peak_name)
    file.close()
    self.current_peak_name_saved='Appended to MatLab peaklist'
    self.showinfo('')
#---------------------------------------------------------------------------------------
  
  def integrate_region(self,spectrum, region, X_axis='w1'):
    '->integrated spectrum along "x" or "y"'

    if spectrum.dimension!=2:
        tkMessageBox.showwarning("Integrate Region","Cannot work with %d dimensions!" % spectrum.dimension)
    
    ppm_per_index = map(lambda w,s: float(w)/s,
                        spectrum.spectrum_width, spectrum.data_size)
#    print ppm_per_index

    #--- assign dimensions
    primary_dimension=dims[X_axis]
    secondary_dimension= not primary_dimension

#    print spectrum.data_height([120,8])

    
    curr_point=[0,0]

    #--- go along main dimension, integrate secondary
    
    integrated=[]

    #--- one index ---    
    min_X=region[0][primary_dimension]
    min_Y=region[0][secondary_dimension]

    max_X=region[1][primary_dimension]
    max_Y=region[1][secondary_dimension]
    
    X_steps=(max_X-min_X)/ppm_per_index[primary_dimension]
    Y_steps=(max_Y-min_Y)/ppm_per_index[secondary_dimension]

    # just for xmgr scaling
    max_y=spectrum.data_height(curr_point)/Y_steps 
    min_y=max_y

    for i in range(X_steps):
      
      curr_X=min_X + ppm_per_index[primary_dimension]*i
      curr_point[primary_dimension]=curr_X
      
      sum=0

      for j in range(Y_steps):

          curr_Y=min_Y + ppm_per_index[secondary_dimension]*j
          curr_point[secondary_dimension]=curr_Y
                     
          height=spectrum.data_height(curr_point)
                     
          sum+=height

      Y_point= sum/Y_steps
      integrated.append([curr_X,Y_point])
                     
      if Y_point>max_y: max_y=Y_point
      if Y_point<min_y: min_y=Y_point

    #--- ranges for plotting ---
    min_x=min_X
    max_x=min_x + ppm_per_index[primary_dimension] * (X_steps-1)
    
    
    return integrated, min_x, max_x, min_y, max_y





#-------------------------------------------------------------
  def fit_integrated(self,integrated_region, fit_func):
    'fit region to two peaks'
    pass
    
  
#######################################################################################
#######################################################################################
#######################################################################################
#######################################################################################

  
# -----------------------------------------------------------------------------
#
class lorentzian_func:
  'lorentzian lineshape peak'

  def parameter_count(self):
    return 3
  
  def value(self, params, x):
    a0, dv, v0 = params
  
    A1=a0*dv*dv
    
    A2=dv*dv-4*(x-v0)*(x-v0)

    return A1/A2


  def parameter_derivatives(self, params, x):

    epsilon=1e-6
    derivatives=[]
    
    for i in range(params):
      #-- step up ---
      temp_params2=params[:]
      temp_params2[i]=params[i]+params[i]*epsilon
      #-- step down ---      
      temp_params1=params[:]
      temp_params1[i]=params[i]-params[i]*epsilon

      
      dfda=(self.value(temp_params2,x)-self.value(temp_params1,x))/params[i]*epsilon
      
      derivatives.append(dfda)

  
    return derivatives

#  

def run(session):

  sputil.the_dialog(region_rmsd_dialog,session).show_window(1)
